@@ -1,7 +1,11 @@
+import re
 import string
 from math import ceil
 from operator import itemgetter
+from random import randrange
 
+import lime
+import lime.lime_tabular
 import matplotlib.pyplot as plt
 import numpy as np
 import pandas as pd
@@ -14,7 +18,8 @@ from sklearn.feature_extraction.text import (
     CountVectorizer,
     TfidfVectorizer,
 )
-from sklearn.linear_model import ElasticNet, LogisticRegression
+from sklearn.linear_model import ElasticNet, LogisticRegression, SGDClassifier
+from sklearn.tree import DecisionTreeClassifier
 from sklearn.metrics import (
     accuracy_score,
     mean_squared_error,
@@ -23,14 +28,18 @@ from sklearn.metrics import (
     r2_score,
     recall_score,
     roc_auc_score,
+    f1_score
 )
 from sklearn.model_selection import train_test_split
+from sklearn.pipeline import Pipeline
 
 from datto.CleanText import CleanText
 
 
 class ModelResults:
-    def most_similar_texts(self, X, num_examples, text_column_name, num_topics=None):
+    def most_similar_texts(
+        self, X, num_examples, text_column_name, num_topics=None, chosen_stopwords=set()
+    ):
         """
         Uses NMF clustering to create n topics based on adjusted word frequencies
 
@@ -63,6 +72,7 @@ class ModelResults:
             | set(["-PRON-"])
             | set(string.punctuation)
             | set([" "])
+            | chosen_stopwords
         )
 
         ct = CleanText()
@@ -242,8 +252,8 @@ class ModelResults:
         med = X_train.median().values.reshape((1, X_train.shape[1]))
         explainer = shap.KernelExplainer(f, med)
         # Runs too slow if X_test is huge, take a representative sample
-        if X_test.shape[0] > 20000:
-            X_test_sample = X_test.sample(20000)
+        if X_test.shape[0] > 1000:
+            X_test_sample = X_test.sample(1000)
         else:
             X_test_sample = X_test
         shap_values = explainer.shap_values(X_test_sample)
@@ -251,6 +261,8 @@ class ModelResults:
         plt.tight_layout()
         plt.savefig(filename)
 
+        return shap_values
+
     def most_common_words_by_group(
         self, X, text_col_name, group_col_name, num_examples, num_times_min, min_ngram,
     ):
@@ -357,7 +369,7 @@ class ModelResults:
         return overall_counts_df
 
     def score_final_model(
-        self, model_type, X_train, y_train, X_test, y_test, full_pipeline
+        self, model_type, X_test, y_test, trained_model, multiclass=False
     ):
         """
         Score your model on the test dataset. Only run this once to get an idea of how your model will perform in realtime.
@@ -366,25 +378,29 @@ class ModelResults:
         Parameters
         --------
         model_type: str
-        X_train: DataFrame
-        y_train: DataFrame
         X_test: DataFrame
         y_test: DataFrame
-        full_pipeline: sklearn Pipeline
+        trained_model: sklearn model
+        multiclass: bool
 
         Returns
         --------
-        full_pipeline: model
+        model: model
             Fit model
         y_predicted: array
         """
-        # Fitting the final model
-        full_pipeline.fit(X_train, y_train)
-
         # Predict actual scores
-        y_predicted = full_pipeline.predict(X_test)
+        y_predicted = trained_model.predict(X_test)
 
-        if model_type.lower() == "classification":
+        if multiclass:
+            pscore = precision_score(y_test, y_predicted, average="weighted")
+            rscore = recall_score(y_test, y_predicted, average="weighted")
+            f1score = f1_score(y_test, y_predicted, average="weighted")
+
+            print(f"Final Model Precision Weighted: {pscore}")
+            print(f"Final Model Recall Weighted: {rscore}")
+            print(f"Final Model F1 Weighted: {f1score}")
+        elif model_type.lower() == "classification":
             pscore = precision_score(y_test, y_predicted)
             rscore = recall_score(y_test, y_predicted)
             accuracy = accuracy_score(y_test, y_predicted)
@@ -414,4 +430,193 @@ class ModelResults:
             print(f"Mean Negative Median Absolute Error: {mae * -1}")
             print(f"Mean R2: {r2}")
 
-        return full_pipeline, y_predicted
+        return trained_model, y_predicted
+
+    def coefficients_summary(
+        self,
+        X,
+        y,
+        num_repetitions,
+        num_coefficients,
+        model_type,
+        multiclass=False,
+        params={},
+    ):
+        """
+        Prints average coefficient values using a regression model.
+
+        Parameters
+        --------
+        X: DataFrame
+        y: DataFrame
+        num_repetitions: int
+            Number of times to create models
+        num_coefficients: int
+            Number of top coefficients to display
+        model_type: str
+            'classification' or 'regression'
+        multiclass: bool
+        params: dict
+            Optional - add to change regression params, otherwise use default
+            
+        Returns
+        --------
+        simplified_df: DataFrame
+            Has mean, median, and standard deviation for coefficients after several runs
+        """
+        coefficients_df = pd.DataFrame(columns=["features", "coefficients"])
+        X["intercept"] = 1
+
+        for _ in range(num_repetitions):
+            if multiclass:
+                model = DecisionTreeClassifier()
+            elif model_type.lower() == "classification":
+                model = LogisticRegression(fit_intercept=False, **params)
+            else:
+                model = ElasticNet(fit_intercept=False, **params)
+
+            X_train, _, y_train, _ = train_test_split(X, y)
+
+            model.fit(X_train, y_train)
+
+            if multiclass:
+                coefs = model.feature_importances_
+            elif model_type.lower() == "classification":
+                coefs = model.coef_[0]
+            else:
+                coefs = model.coef_
+
+            temp_df = pd.DataFrame(
+                [x for x in zip(X_train.columns, coefs)],
+                columns=["features", "coefficients"],
+            )
+
+            coefficients_df = coefficients_df.append(temp_df)
+
+        column_of_interest = "coefficients"
+
+        summary_coefficients_df = pd.DataFrame(
+            coefficients_df.groupby("features").agg(
+                {column_of_interest: ["mean", "std", "median"]}
+            )
+        )
+
+        summary_coefficients_df.columns = ["mean", "std", "median"]
+
+        summary_coefficients_df.reset_index(inplace=True)
+
+        value_counts_df = pd.DataFrame(columns=["features"])
+        for col in X_train.columns:
+            temp_df = pd.DataFrame([[col,]], columns=["features"],)
+            value_counts_df = value_counts_df.append(temp_df)
+        value_counts_df.reset_index(inplace=True, drop=True)
+
+        combined_df = summary_coefficients_df.merge(
+            value_counts_df, on="features", how="left"
+        )
+
+        combined_df["abs_val_mean"] = abs(combined_df["mean"])
+        combined_df["abs_val_to_se"] = abs(combined_df["mean"]) / combined_df["std"]
+
+        combined_df.sort_values("abs_val_to_se", inplace=True, ascending=False)
+
+        simplified_df = (
+            combined_df[["features", "mean", "std", "median", "abs_val_to_se"]]
+            .head(num_coefficients)
+            .round(7)
+        )
+
+        # 7 is the most before it flips back to scientific notation
+        print("Coefficients summary (descending by mean abs se value):")
+        print(simplified_df)
+
+        return simplified_df
+
+    def coefficients_individual_predictions(
+        self,
+        trained_model,
+        X_train,
+        X_test,
+        id_col,
+        num_samples,
+        model_type,
+        class_names=["False", "True"],
+    ):
+        def model_preds_adjusted(data):
+            if model_type.lower() == "classification":
+                predictions = np.array(trained_model.predict_proba(data))
+            else:
+                predictions = np.array(trained_model.predict(data))
+            return predictions
+
+        if model_type.lower() == "classification":
+            explainer = lime.lime_tabular.LimeTabularExplainer(
+                np.array(X_train),
+                feature_names=X_train.columns,
+                class_names=class_names,
+                mode="classification",
+            )
+        else:
+            explainer = lime.lime_tabular.LimeTabularExplainer(
+                np.array(X_train),
+                feature_names=X_train.columns,
+                class_names=class_names,
+                mode="regression",
+            )
+
+        for i in range(num_samples):
+            user_idx = randrange(X_test.shape[0])
+            exp = explainer.explain_instance(
+                np.array(X_test.iloc[user_idx]),
+                model_preds_adjusted,
+                num_features=50,
+                top_labels=10,
+            )
+
+            user_id = X_test.iloc[user_idx][id_col]
+            print(f"\nUser: {user_id}")
+
+            if model_type.lower() == "classification":
+                prediction = class_names[
+                    trained_model.predict_proba(
+                        pd.DataFrame(X_test.iloc[user_idx]).T
+                    ).argmax()
+                ]
+            else:
+                prediction = round(
+                    trained_model.predict(pd.DataFrame(X_test.iloc[user_idx]).T)[0], 7
+                )
+
+            print('We predicted "{}" for this user because...\n'.format(prediction))
+            features_list = []
+            for feature in exp.as_list():
+                features_list.append({user_id: exp.as_list()})
+                try:
+                    feature_name = re.findall("<.*<|>.*>", feature[0])[0]
+                except Exception:
+                    feature_name = re.findall(".*<|.*>", feature[0])[0]
+                cleaned_feature_name = (
+                    feature_name.replace("<=", "")
+                    .replace(">=", "")
+                    .replace("<", "")
+                    .replace(">", "")
+                    .strip()
+                )
+                comparison_val = float(
+                    feature[0].split(" <")[-1].split(" >")[-1].split("=")[-1].strip()
+                )
+                actual_feature_val = X_test.iloc[user_idx][cleaned_feature_name]
+                last_operator = feature[0].split(" ")[-2]
+
+                if "<" in last_operator and actual_feature_val <= comparison_val:
+                    print(" * This is true: " + feature[0])
+                else:
+                    print(" * This is false: " + feature[0])
+            exp.as_pyplot_figure()
+            plt.tight_layout()
+            plt.savefig(f"lime_graph_user_{user_id}.png")
+
+            print("\n\n~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n")
+
+        return features_list
+
